def krange(range_min_value, range_max_value):
    def decorator(func):
        def wrapper(*args, **kwargs):
            for arg_value in args:
                if isinstance(arg_value, (int, float)) and (
                    arg_value < range_min_value or arg_value > range_max_value
                ):
                    raise ValueError(
                        f"Value should be in the {range_min_value}-{range_max_value} range"
                    )
            return func(*args, **kwargs)

        return wrapper

    return decorator



from kaya_module_sdk.src.utils.metadata.abstract_metadata import KMetadata
from kaya_module_sdk.src.utils.metadata.abstract_validation import KValidation


class MinLen(KMetadata, KValidation):
    def __init__(self, value: float) -> None:
        self._data = {
            "min_len": value,
        }

    def __str__(self) -> str:
        return f'minlen:{self._data["min_len"]}'

    def load(self, str_repr: str) -> dict:
        segmented = str_repr.split(":")
        if not len(segmented) == 2 or segmented[0] != "minlen":
            return {}
        self._data["min_len"] = float(segmented[1])
        return self._data



from logging import Logger, getLogger
import importlib
import inspect
import heapq

from abc import ABC, abstractmethod
from typing import Annotated, Any, get_args, get_origin, get_type_hints

import pysnooper  # type: ignore

from kaya_module_sdk.src.module.arguments import Args
from kaya_module_sdk.src.module.config import KConfig
from kaya_module_sdk.src.module.returns import Rets
from kaya_module_sdk.src.utils.metadata.abstract_metadata import KMetadata
from kaya_module_sdk.src.utils.metadata.abstract_validation import KValidation
from kaya_module_sdk.src.utils.metadata.display_description import DisplayDescription
from kaya_module_sdk.src.utils.metadata.display_name import DisplayName

log: Logger = getLogger(__name__)


class Config(KConfig):
    def __init__(self):
        super().__init__()


class Module(ABC):
    """[ DESCRIPTION ]: Kaya Strategy Module Template.

    [ MANIFEST ]: {
        "moduleName": "string",         -- Name of the module
        "moduleDisplayLabel": "string", -- Display label for this module in the frontend
        "moduleCategory": "enum",       -- Category of the module. An ENUM defined by the NeptuneAPI smithy model.
        "moduleDescription": "string",  -- Description of the module
        "author": "kaya_id(user)",      -- UserID of the User vertex that owns this module.
        "inputs": [{
            "name": "string",           -- Name of the input field in the request object
            "label": "string",          -- Display label for this input in the frontend
            "type": "kaya_id(value)",   -- VertexID of the value that represents this input datatype
            "description": "string",    -- Description of the INPUT
            "validations": [
                "validation_pattern"    -- An array of validation queries to run against the inputs.
            ]}
        ],
        "outputs": [{
            "name": "string",           -- Name of the output field in the returned structure
            "label": "string",          -- Display label for this output in the frontend
            "type": "kaya_id(value)",   -- VertexID of the value that represents this output datatype
            "description": "string"     -- Description of the output
            "validations": [
                "validation_pattern"    -- An array of validation queries to run against the inputs.
            ]}
        ]
    }
    """

    config: KConfig
    subclasses: list = []
    modules: dict = {}
    _manifest: dict = {}
    _recompute_manifest: bool = True

    def __init__(self) -> None:
        self.config = Config()
        self.import_subclasses()
        self.modules = {item.__class__.__name__: item for item in self.subclasses}

    @pysnooper.snoop()
    def import_subclasses(self) -> list:
        module_name = self.__module__
        log.info("Importing package %s", module_name)
        package = importlib.import_module(module_name).__package__
        log.info("Searching package for subclasses at %s.module", package)
        if not package:
            return self.subclasses
        try:
            module = importlib.import_module(f"{package}.module")
        except TypeError:
            return self.subclasses
        for cls_name, obj in inspect.getmembers(module):
            if (
                inspect.isclass(obj)
                and issubclass(obj, Module)
                and obj not in [Module, Args, Rets, KConfig, KMetadata, KValidation]
                and cls_name != "KayaStrategyModule"
            ):
                subclass_instance = obj()
                self.subclasses.append(subclass_instance)
        return self.subclasses

    @pysnooper.snoop()
    def _extract_manifest(self) -> dict[str, Any]:
        main_method = self.main
        type_hints = get_type_hints(main_method)
        args_hints = {
            param: type_hint
            for param, type_hint in type_hints.items()
            if param != "return"
        }
        return_hint = type_hints.get("return", None)
        signature = inspect.signature(main_method)
        params_metadata = {}
        for param_name, param in signature.parameters.items():
            expected_type = args_hints.get(param_name, "Any")
            params_metadata[param_name] = {
                "expected_type": expected_type,
                "annotation": param.annotation,
                "details": self._get_class_metadata(expected_type)
                if inspect.isclass(expected_type)
                else None,
            }
        return_metadata = {
            "expected_type": return_hint,
            "annotation": signature.return_annotation,
            "details": self._get_class_metadata(return_hint)
            if inspect.isclass(return_hint)
            else None,
        }
        metadata = {
            "moduleName": str(self.config.name),
            "moduleVersion": str(self.config.version),
            "moduleDisplayLabel": str(self.config.display_label),
            "moduleCategory": str(self.config.category),
            "moduleDescription": str(self.config.description),
            "author": str(self.config.author),
            "inputs": [],
            "outputs": [],
        }
        metadata["inputs"] += self._order_records_by_priority(
            *self._extract_metadata(params_metadata["args"]["details"])
        )
        metadata["outputs"] += self._order_records_by_priority(
            *self._extract_metadata(return_metadata["details"])
        )
        return metadata

    @pysnooper.snoop()
    def _extract_metadata(self, details: dict) -> list[dict[str, str]]:
        metadata: list[dict[str, str]] = []
        if not details:
            return metadata
        for detail in details:
            unpacked = self._unpack_annotated(details[detail]["type"])
            record = {
                "name": detail.strip("_"),
                "label": detail,
                "type": str(unpacked[0]).split("'")[1],
                "description": None,
                "validations": [],
            }
            for item in unpacked[1]:
                if not isinstance(item, KMetadata):
                    continue
                segmented = str(item).split(":")
                if isinstance(item, DisplayName):
                    record["label"] = segmented[1]
                elif isinstance(item, DisplayDescription):
                    record["description"] = segmented[1]
                elif isinstance(item, KValidation):
                    record["validations"].append(str(item))
            metadata.append(record)
        return metadata

    @pysnooper.snoop()
    def _order_records_by_priority(self, *records: dict):
        ordered: list = []
        leftover: list = []
        to_order: list = []
        priority_queue: list = []

        for record in records:
            position = [
                int(item.split(":")[1])
                for item in record.get("validations", [])
                if item.startswith("position:")
            ]
            if not position:
                leftover.append(record)
                continue
            to_order.append(
                (
                    position[0],
                    record,
                )
            )
        for record in to_order:
            heapq.heappush(priority_queue, record)
        while priority_queue:
            _, record = heapq.heappop(priority_queue)
            ordered.append(record)
        ordered += leftover
        return ordered

    @pysnooper.snoop()
    def _unpack_annotated(self, annotated_type):
        # Check if the type is an Annotated type
        if get_origin(annotated_type) is Annotated:
            base_type, *metadata = get_args(annotated_type)
            # Check if the base type is another Annotated type
            if get_origin(base_type) is Annotated:
                return self._unpack_annotated(base_type)
            return base_type, metadata
        # If the type is a list, we need to unpack its inner type
        elif get_origin(annotated_type) is list:
            inner_type = get_args(annotated_type)[0]
            return self._unpack_annotated(inner_type)
        return annotated_type, []

    @pysnooper.snoop()
    def _get_class_metadata(self, cls: type) -> dict[str, Any]:
        """Recursively fetch the metadata for class attributes with type
        annotations."""
        if not hasattr(cls, "__annotations__"):
            return {}
        class_metadata = {}
        for attr_name, attr_type in cls.__annotations__.items():
            class_metadata[attr_name] = {
                "type": attr_type,
                "details": self._get_class_metadata(attr_type)
                if inspect.isclass(attr_type)
                else None,
            }
        return class_metadata

    @property
    def manifest(self) -> dict:
        """
        [ RETURN ]: {
            "moduleName": "string",
            "moduleVersion": "2.0",
            "moduleDisplayLabel": "string",
            "moduleCategory": "enum",
            "moduleDescription": "string",
            "author": "kaya_id(user)",
            "inputs": [
                {
                "name": "string",
                "label": "string",
                "type": "kaya_id(value)",
                "description": "string",
                "validations": [
                    "validation_pattern"
                ]
                }
            ],
            "outputs": [
                {
                "name": "string",
                "label": "string",
                "type": "kaya_id(value)",
                "description": "string"
                }
            ]
        }
        """
        if self._manifest and not self._recompute_manifest:
            return self._manifest
        self._manifest = self._extract_manifest()
        return self._manifest

    @abstractmethod
    def main(self, args: Args) -> Rets:
        pass


# CODE DUMP



from kaya_module_sdk.src.utils.metadata.abstract_metadata import KMetadata
from kaya_module_sdk.src.utils.metadata.abstract_validation import KValidation


class GTE(KMetadata, KValidation):
    def __init__(self, value: float) -> None:
        self._data = {
            "greater_than_or_equal_to": value,
        }

    def __str__(self) -> str:
        return f'>=:{self._data["greater_than_or_equal_to"]}'

    def load(self, str_repr: str) -> dict:
        segmented = str_repr.split(":")
        if not len(segmented) == 2 or segmented[0] != ">=":
            return {}
        self._data["greater_than_or_equal_to"] = float(segmented[1])
        return self._data



import json
from logging import Logger, getLogger
import os

import pysnooper  # type: ignore
from kaya_module_sdk.src.exceptions.malformed_results import MalformedResultsException
from kaya_module_sdk.src.exceptions.write_failure import WriteFailureException

log: Logger = getLogger(__name__)


class KVLR:
    """[ KVL(R)eporter ]: Responsibilities -

    * Processes and formats results received from KVL(E)
    * Generates JSON report dump file if specified
    """

    report: dict
    context: dict
    dump_file_path: str

    def __init__(self, dump_file: str | None = None, **kwargs: dict) -> None:
        self.report = {}
        self.context = kwargs
        self.dump_file_path = dump_file or "kvl.report"

    # UTILS

    @pysnooper.snoop()
    def _check_dump_file_path(self, *args, **kwargs) -> bool:
        directory = os.path.dirname(self.dump_file_path)
        if not directory:
            directory = "."
        if (
            not os.path.exists(directory)
            or not os.path.isdir(directory)
            or not os.access(directory, os.W_OK)
        ):
            raise WriteFailureException(
                "KVL(R)eporter dump file {self.dump_file_path} has an unexistent "
                "parent directory or location restricts write permissions."
            )
        return True

    @pysnooper.snoop()
    def _check_kvle_results(self, *results, **kwargs):
        if not results:
            raise MalformedResultsException(
                "KVL(R)eporter received malformed results to process!"
            )
        for result in results:
            if (
                "source" not in result
                and "meta" not in result
                and "rules" not in result
            ):
                raise MalformedResultsException(
                    "KVL(R)eporter received malformed results to process!"
                )
        return True

    @pysnooper.snoop()
    def check_preconditions(self, *results: dict, dump: bool = False) -> bool:
        check_results, preconditions = {}, {
            "dump_file_path": self._check_dump_file_path,
            "kvle_results": self._check_kvle_results,
        }
        for check, func in preconditions.items():
            if check == "dump_file_path" and not dump:
                continue
            check_results.update({check: func(*results)})
        return True

    @pysnooper.snoop()
    def _dump_test_results_report_json(self, formatted_results, file_path):
        try:
            with open(file_path, "ab") as json_file:
                json.dump(formatted_results, json_file, indent=4)
            print(f"[ INFO ]: KVL(R) test reports dumped to: {file_path}")
        except (IOError, TypeError) as e:
            raise WriteFailureException(
                "An error occurred while writing to the file"
            ) from e
        return True

    # FORMATTERS

    def _format_validation_result(self, result: dict) -> list:
        builder = []
        for k in result:
            validation_report = {}
            if k not in ("source", "rules", "meta"):
                continue
            if k == "source":
                validation_report = {
                    "Source File Check": {
                        "FILES": result[k]["ok"] + result[k]["nok"],
                        "RESULT": "NOK" if result[k]["nok"] else "OK",
                    }
                }
            elif k == "meta":
                validation_report = {
                    "Module Metadata Check": {
                        "MODULES": result[k]["ok"] + result[k]["nok"],
                        "RESULT": "NOK" if result[k]["nok"] else "OK",
                    }
                }
            elif k == "rules":
                validation_report = {
                    "Module Constraint Rules Check": {
                        "MODULES": result[k]["ok"] + result[k]["nok"],
                        "RESULT": "NOK" if result[k]["nok"] else "OK",
                    }
                }
            builder.append(validation_report)
        return builder

    def _format_validation_report(self, *results: dict) -> list:
        builder = []
        for result_dict in results:
            builder += self._format_validation_result(result_dict)
        return builder

    # ACTIONS

    @pysnooper.snoop()
    def generate_report(self, *results: dict, dump: bool = False):
        """[ INPUT ]:

        [ RETURN ]:
        """
        self.check_preconditions(*results, dump=dump)
        formatted = self._format_validation_report(*results)
        if dump:
            self._dump_test_results_report_json(formatted, self.dump_file_path)
        return formatted


# CODE DUMP

#           'Source File Check': {
#               "FILES": [
#                   {
#                       "path": "./src/modules/mymodule.py",
#                       "errors": [],
#                       "result": "OK"
#                   },
#                   {
#                       "path": "./src/modules/myothermoduleIforgotabout.py",
#                       "errors": [
#                           {
#                               "line": 13,
#                               "msg": "We don't do that here",
#                               "code": "KVL-11"
#                           }
#                       ]
#                       "result": "NOK",
#                   }
#               ],
#               "RESULT": "NOK",
#           },

#           'Module Metadata': {
#               "MODULES": {
#                   "Dummy1": {
#                       "package": 'dummy_module',
#                       "meta": {
#                           "VERSION": {
#                               "required": true,
#                               "set": true,
#                               "valid": true,
#                               "value": "1.0.1"
#                           },
#                           "CATEGORY": {
#                               "required": true,
#                               "set": true,
#                               "valid": true,
#                               "value": "INDICATOR"
#                           },
#                           "TYPE": {
#                               "required": true,
#                               "set": true,
#                               "valid": false,
#                               "value": "NotAValidType"
#                           }
#                       },
#                       "errors": ["TYPE"],
#                       "result": "NOK"
#                   }
#               },
#               "RESULT": "NOK"
#           },

#           'Module Constraint Rules Check': {
#               "MODULES": {
#                   "Dummy1": {
#                       "package": 'dummy_module',
#                       "function": "main",
#                       "rules": [
#                           {
#                               "name": "Large Window Rule",
#                               "verb": "gte",
#                               "args": ["window", "data.length"]
#                           },
#                           {
#                               "name": "Single Data Point",
#                               "verb": "eq",
#                               "rule": ["window", 1]
#                           }
#                       ],
#                       "errors": [],
#                       "result": "OK"
#                   }
#               },
#               "RESULT": "OK"
#           },
#       }


#           - *results: [{
#               'meta': {
#                   'ok': [{
#                       'package': 'dummy_package',
#                       'module': 'Dummy1',
#                       'version': {
#                           'required': True,
#                           'set': True,
#                           'valid': True,
#                           'value': '1.0.1',
#                       },
#                       'category': {
#                           'required': True,
#                           'set': True,
#                           'valid': True,
#                           'value': 'INDICATOR',
#                       },
#                       'type': {
#                           'required': True,
#                           'set': True,
#                           'valid': False,
#                           'value': 'NotAValidType'
#                       },
#                   }],
#                   'nok': {}

#               'source': [{
#                   'path': './src',
#                   'errors': [],
#                   'result': 'OK',
#               }],

#               'rules': [{
#                   'package': 'dummy_package',
#                   'module': 'Dummy1',
#                   'functions': {
#                       'main': [{
#                           'name': 'Large Window Rule',
#                           'verb': 'gte',
#                           'rule': ['window', 'data.length']
#                       },
#                       {
#                           'name': 'Single Data Point',
#                           'verb': 'eq',
#                           'rule': ['single_window', 1],
#                       }],
#                   },
#               }],
#               }]



def klt(lt_value):
    def decorator(func):
        def wrapper(*args, **kwargs):
            for arg_value in args:
                if isinstance(arg_value, (int, float)) and arg_value >= lt_value:
                    raise ValueError(f"Value should be less than to {lt_value}")
            return func(*args, **kwargs)

        return wrapper

    return decorator



import importlib
import logging
import os

import pysnooper  # type: ignore
from kaya_module_sdk.src.exceptions.kvl_failure import KVLFailureException
from kaya_module_sdk.src.testing.kvl_executer import KVLE
from kaya_module_sdk.src.testing.kvl_reporter import KVLR

log = logging.getLogger(__name__)


class KVL:
    """[ KVL(H)arness ]: Responsibilities -

    * Searches Python3 source code files
        [ NOTE ]: If a directory is specified, looks for all *.py files
    * Runs KVL(E) with source file paths and/or module names depending on action
        When validating source code files it will receive the paths of the given files as strings
        When validating module input data at runtime it will receive the module name.
    * Runs KVL(R) with KVL(E) data
    * Maintains a stable interface for KVL's backend
    """

    _python_files: dict
    _python_module: str
    _filename_convention: dict
    _module_package: dict
    _kvl: dict

    def __init__(self, *args, **kwargs) -> None:
        self._filename_convention = {"suffix": ".py"}
        self._module_package = {
            "name": kwargs.get("module_name"),
            "instance": None,
        }
        self._python_files = {fl_path: [] for fl_path in args}
        self._kvl = {
            "executer": None,
            "reporter": None,
        }

    # UTILS

    @pysnooper.snoop()
    def _search_python_files_in_dir(self, dir_path: str) -> list:
        python_files = []
        for root, _, files in os.walk(dir_path):
            for fl in files:
                if fl.endswith(self._filename_convention["suffix"]):
                    python_files.append(os.path.join(root, fl))
        return python_files

    @pysnooper.snoop()
    def _import_package(self, package_name: str):
        pkg = importlib.import_module(package_name)
        return pkg

    @pysnooper.snoop()
    def _load_python_files(self, target_files: list[str]) -> dict:
        for file_path in target_files:
            with open(file_path, encoding="utf-8") as fl:
                data = fl.readlines()
            self._python_files.update({file_path: data})
        return self._python_files

    @pysnooper.snoop()
    def _load_module_package(self, module_name: str) -> dict:
        loaded_pkg = self._import_package(f'{module_name.replace("-", "_")}.module')
        class_ = getattr(loaded_pkg, "KayaStrategyModule")
        instance = class_()
        self._module_package.update({"name": module_name, "instance": instance})
        return self._module_package

    @pysnooper.snoop()
    def _get_kvl_executer(self, **kwargs: dict):
        if not self._kvl["executer"]:
            self._kvl["executer"] = KVLE(**kwargs)
        return self._kvl["executer"]

    @pysnooper.snoop()
    def _get_kvl_reporter(self, **kwargs: dict):
        if not self._kvl["reporter"]:
            self._kvl["reporter"] = KVLR(**kwargs)  # type: ignore
        return self._kvl["reporter"]

    # ACTIONS

    @pysnooper.snoop()
    def check_rules(
        self,
        module: str | None = None,
        dump_report: bool = False,
    ) -> dict:
        """[ NOTE ]: KVL(H) entry point for module constraints
        verifications."""
        if not module:
            raise KVLFailureException("No module package name specified!")
        module_data = self._load_module_package(module)
        try:
            verification_results = self._get_kvl_executer(**self.__dict__).check_rules(
                module_data
            )
            report = self._get_kvl_reporter(**self.__dict__).generate_report(
                {"rules": verification_results}, dump=dump_report
            )
        except Exception as e:
            raise KVLFailureException("Module metadata verification failed!") from e
        return report

    @pysnooper.snoop()
    def check_meta(
        self,
        module: str | None = None,
        dump_report: bool = False,
    ) -> dict:
        """[ NOTE ]: KVL(H) entry point for module metadata verifications."""
        if not module:
            raise KVLFailureException("No module package name specified!")
        module_data = self._load_module_package(module)
        try:
            verification_results = self._get_kvl_executer(**self.__dict__).check_meta(
                module_data
            )
            report = self._get_kvl_reporter(**self.__dict__).generate_report(
                {"meta": verification_results}, dump=dump_report
            )
        except Exception as e:
            raise KVLFailureException(
                "Module metadata verification failed! Details"
            ) from e
        return report

    @pysnooper.snoop()
    def check_source(
        self,
        file_path: str | None = None,
        dump_report: bool = False,
    ) -> dict:
        """[ NOTE ]: KVL(H) entry point for source code verifications."""
        target_path = file_path or "."
        python_file_paths = (
            self._search_python_files_in_dir(target_path)
            if os.path.isdir(target_path)
            else [target_path]
        )
        loaded_files = self._load_python_files(python_file_paths)
        try:
            verification_results = self._get_kvl_executer(**self.__dict__).check_source(
                loaded_files
            )
            report = self._get_kvl_reporter(**self.__dict__).generate_report(
                {"source": verification_results}, dump=dump_report
            )
        except Exception as e:
            raise KVLFailureException("Source file verification failed!") from e
        return report

    @pysnooper.snoop()
    def check(self, *targets: str, **kwargs: dict) -> dict:
        """[ INPUT ]:

            * Positional arguments:

            - *targets - type string, options (meta|source|rules|all)

            * Keyword arguments:

            - report= - type bool, format and display verification results
            - dump_report= - type bool, implies <report=True> - dumps report to file
            - module= - type str, implies <(meta|rules|all) in *targets> - module package name
            - file_path= - type str, implies <(source|all) in *targets> - python file or directory to check recursively

        [ RETURN ]: KVL(R) report - type dict

        [ NOTE ]: Main KVL(H) entry point/interface for all verification types.
        """
        checkers = {
            "meta": self.check_meta,
            "source": self.check_source,
            "rules": self.check_rules,
        }
        to_check = checkers.keys() if "all" in targets else list(targets)
        if not to_check:
            log.error("No verification targets specified!")
            return {}
        check_results = {target: checkers[target](**kwargs) for target in to_check}
        return check_results


# CODE DUMP



from importlib.util import find_spec
from logging import Logger, getLogger

import mypy.api as mypy_api
import pysnooper  # type: ignore
from flake8.api import legacy as flake8  # type: ignore
from kaya_module_sdk.src.exceptions.module_not_found import ModuleNotFoundException
from kaya_module_sdk.src.utils.metadata.equal import EQ
from kaya_module_sdk.src.utils.metadata.greater import GT
from kaya_module_sdk.src.utils.metadata.greater_or_equal import GTE
from kaya_module_sdk.src.utils.metadata.less import LT
from kaya_module_sdk.src.utils.metadata.less_or_equal import LTE
from kaya_module_sdk.src.utils.metadata.max_len import MaxLen
from kaya_module_sdk.src.utils.metadata.maximum import Max
from kaya_module_sdk.src.utils.metadata.min_len import MinLen
from kaya_module_sdk.src.utils.metadata.minimum import Min
from kaya_module_sdk.src.utils.metadata.value_range import ValueRange

log: Logger = getLogger(__name__)


class KVLE:
    """[ KVL(E)xecuter ]: Responsibilities -

    * Runs all linters and checkers on the source code files
        [ Ex ]: pylint, flake8, mypy
    * Imports all modules found in source code files
    * Attempts extraction of Manifest from imported modules
    * Verifies that required module metadata is properly set
    * Checks module specific input data validators are properly set
    * Collects all validation results in a easy to handle manner
    """

    check: dict[str, dict[str, list]]
    context: dict

    def __init__(self, **kvlh_context):
        self.context = kvlh_context
        self.check = {
            "rules": {
                "ok": [],
                "nok": [],
            },
            "meta": {
                "ok": [],
                "nok": [],
            },
            "source": {
                "ok": [],
                "nok": [],
            },
        }

    # UTILS

    @pysnooper.snoop()
    def check_package_installed(self, test: dict) -> bool:
        try:
            find_spec(test["package"])
        except ModuleNotFoundError as e:
            raise ModuleNotFoundException(
                f'Package {test["package"]} is not installed!'
            ) from e
        return True

    @pysnooper.snoop()
    def load_constraint_rules(self, *rules: str) -> dict:
        matches, metadata_classes = [], [
            EQ,
            GT,
            GTE,
            LT,
            LTE,
            MaxLen,
            Max,
            MinLen,
            Min,
            ValueRange,
        ]
        for rule in rules:
            for cls in metadata_classes:
                try:
                    if ";" in rule:
                        instance = cls(None, None)
                    else:
                        instance = cls(None)
                    instance.load(rule)
                except Exception:
                    continue
                matches.append(instance)
        formatted_matches = {
            item.__class__.__name__: item
            for item in matches
            if None not in item._data.values()
        }
        return {} if rules and not formatted_matches else formatted_matches

    # ACTIONS

    @pysnooper.snoop()
    def check_rules(self) -> dict:
        """
        [ RETURN ]: {
            'ok': [{
                'package': 'dummy_package',
                'module': 'Dummy1',
                'functions': {
                    'main': [{
                        'name': 'Large Window Rule',
                        'verb': 'gte',
                        'rule': ['window', 'data.length']
                    },
                    {
                        'name': 'Single Data Point',
                        'verb': 'eq',
                        'rule': ['single_window', 1],
                    }],
                },
            }],
            'nok': [],
        }
        """
        package = self.context["_module_package"]["instance"]
        package_name = package.config.name
        submodules = package.modules
        module_report: dict = {}
        for module_name in submodules:
            nok_flag, module_record = False, {
                "package": package_name,
                "module": module_name,
                "functions": {"main": []},
            }
            manifest = submodules[module_name].manifest
            if not manifest:
                nok_flag = True
                module_report.update({"manifest": {"required": True, "set": False}})
                continue
            if (
                manifest
                and not isinstance(manifest, dict)
                or not manifest.get("inputs")
                or not manifest.get("outputs")
            ):
                nok_flag = True
                module_report.update(
                    {
                        "manifest": {
                            "required": True,
                            "set": True,
                            "valid": False,
                            "value": manifest,
                        }
                    }
                )
                continue
            for arg in manifest["inputs"]:
                if not arg["validations"]:
                    continue
                loaded = self.load_constraint_rules(*arg["validations"])
                if arg["validations"] and not loaded:
                    nok_flag = True
                for constraint_name in loaded:
                    values = list(loaded[constraint_name]._data.values())
                    if len(values) == 1:
                        values = values[0]
                    module_record["functions"]["main"].append(
                        {
                            "name": constraint_name,
                            "target": "inputs",
                            "verb": constraint_name.lower(),
                            "field": arg["label"],
                            "rule": [constraint_name, values],
                        }
                    )
            for ret in manifest["outputs"]:
                if not ret["validations"]:
                    continue
                loaded = self.load_constraint_rules(*ret["validations"])
                if ret["validations"] and not loaded:
                    nok_flag = True
                for constraint_name in loaded:
                    values = list(loaded[constraint_name]._data.values())
                    if len(values) == 1:
                        values = values[0]
                    module_record["functions"]["main"].append(
                        {
                            "name": constraint_name,
                            "target": "outputs",
                            "verb": constraint_name.lower(),
                            "field": arg["label"],
                            "rule": [constraint_name, values],
                        }
                    )
            if nok_flag:
                self.check["rules"]["nok"].append(module_record)
                continue
            self.check["rules"]["ok"].append(module_record)
        return self.check["rules"]

    # @pysnooper.snoop()
    def check_meta(self, module_data: dict, report: bool = True) -> dict:
        """
        [ RETURN ]: {
            'ok': [],
            'nok': [{
                'package': 'dummy_package',
                'module': 'Dummy1',
                'version': {
                    'required': True,
                    'set': True,
                    'valid': True,
                    'value': '1.0.1',
                },
                'category': {
                    'required': True,
                    'set': True,
                    'valid': True,
                    'value': 'INDICATOR',
                },
                'type': {
                    'required': True,
                    'set': True,
                    'valid': False,
                    'value': 'NotAValidType'
                },
            }],
        }
        """
        package = self.context["_module_package"]["instance"]
        package_name, package_version = package.config.name, package.config.version
        submodules = package.modules
        for module_name in submodules:
            nok_flag, module_report = False, {
                "package": package_name,
                "package_version": package_version,
                "module": module_name,
            }
            if not submodules[module_name].config:
                nok_flag = True
                module_report.update({"config": {"required": True, "set": False}})
            for key in submodules[module_name].config._mandatory:
                if not submodules[module_name].config.__dict__.get(key):
                    nok_flag = True
                    module_report.update({key: {"required": True, "set": False}})
                    continue
                if (
                    key != "version"
                    and not type(submodules[module_name].config.__dict__[key])
                    not in (str,)
                ) or (
                    key == "version"
                    and type(submodules[module_name].config.__dict__[key])
                    not in (str, float)
                ):
                    nok_flag = True
                    module_report.update(
                        {
                            key: {
                                "required": True,
                                "set": True,
                                "valid": False,
                                "value": submodules[module_name].config.__dict__[key],
                            }
                        }
                    )
                    continue
                module_report.update(
                    {
                        key: {
                            "required": True,
                            "set": True,
                            "valid": True,
                            "value": submodules[module_name].config.__dict__[key],
                        }
                    }
                )

            manifest = submodules[module_name].manifest
            if not manifest:
                nok_flag = True
                module_report.update(
                    {
                        "manifest": {
                            "required": True,
                            "set": False,
                            "valid": False,
                            "value": manifest,
                        }
                    }
                )

            if manifest and not isinstance(manifest, dict):
                nok_flag = True
                module_report.update(
                    {
                        "manifest": {
                            "required": True,
                            "set": True,
                            "valid": False,
                            "value": manifest,
                        }
                    }
                )
            if nok_flag:
                self.check["meta"]["nok"].append(module_report)
                continue
            self.check["meta"]["ok"].append(module_report)
        return self.check["meta"]

    # @pysnooper.snoop()
    def check_source(self, loaded_files: dict, report=True) -> dict:
        """
        [ RETURN ]: {
            'ok': [{
                'path': './src',
                'errors': [],
                'result': 'OK',
            }],
            'nok': [],
        }
        """
        for file_path, file_lines in loaded_files.items():
            errors = []
            style_guide = flake8.get_style_guide()
            flake8_report = style_guide.check_files([file_path])
            if flake8_report.total_errors > 0:
                flake8_output = [
                    {
                        error[0]: [error[1], error[2]]
                        for error in flake8_report._application.file_checker_manager.results
                    }
                ]
                errors.append({"tool": "flake8", "output": flake8_output})
            mypy_result = mypy_api.run([file_path])
            mypy_stdout, mypy_stderr, mypy_exit_status = mypy_result
            if mypy_stderr or mypy_exit_status:
                errors.append(
                    {
                        "tool": "mypy",
                        "output": mypy_stdout + mypy_stderr,
                        "exit": mypy_exit_status,  # type: ignore
                    }
                )
            if errors:
                self.check["source"]["nok"].append(
                    {"path": file_path, "errors": errors, "result": "NOK"}
                )
            else:
                self.check["source"]["ok"].append(
                    {"path": file_path, "errors": [], "result": "OK"}
                )
        return self.check["source"]


# CODE DUMP



import logging
from abc import ABC
from typing import Annotated, get_type_hints

from kaya_module_sdk.src.utils.metadata.display_description import DisplayDescription
from kaya_module_sdk.src.utils.metadata.display_name import DisplayName

log = logging.getLogger(__name__)


class Args(ABC):
    _errors: Annotated[
        list,
        DisplayName("Errors"),
        DisplayDescription("Collection of things that went very, very wrong."),
    ]

    def __init__(self):
        self._errors = []

    @property
    def errors(self) -> list[Exception]:
        return self._errors

    def set_errors(self, *values: Exception) -> None:
        self._errors += list(values)

    def metadata(self):
        return get_type_hints(self, include_extras=True)



def kmaxlen(max_len_value):
    def decorator(func):
        def wrapper(*args, **kwargs):
            if len(args) - 1 > max_len_value:
                raise ValueError(
                    f"Composite type length should not be above {max_len_value}"
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator
